// code to solve exercises for IMPRS topology course

#include <iostream>
#include <iomanip>
#include <fstream>
#include <complex>
#include <vector>
#include <assert.h>
#include <math.h>
#include "matrix.h"
#ifndef NO_MPI
    #include <mpi.h>
#endif

//physics parameters
#define NUMK    100
#define NORB    100 // standard matrix size; 50 chains times 2 (Pauli matrix dimension)
#define THOP    1.0
#define MU      -2.0
#define DELTA   0.3
#define GAMMA   0.5

using namespace std;

typedef complex<double> cdouble;
typedef matrix<cdouble> cmat;
cdouble II(0,1);



// LAPACK stuff
extern "C" void zheev_(char* jobz, char* uplo, int* N, cdouble* H, int* LDA, double* W, cdouble* work, int* lwork, double* rwork, int *info);

int     matsize = NORB;
int     lwork = 2*(2*NORB-1);
double  rwork[3*NORB-2];
cdouble work[2*(2*NORB-1)];
char    jobz = 'V';
char    uplo = 'U';
int     info;


void diagonalize(cmat &H, vector<double> &evals)
{
	zheev_(&jobz, &uplo, &matsize, H.ptr(), &matsize, &evals[0], &work[0], &lwork, &rwork[0], &info);
	assert(!info);
}

void times(cmat &A, cmat &B, cmat &C)
{
    for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += A(i,k)*B(k,j);
            }
        }
}

void times_dn(cmat &A, cmat &B, cmat &C)
{
	for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += conj(A(k,i))*B(k,j);
            }
        }
}

void times_nd(cmat &A, cmat &B, cmat &C)
{
	for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += A(i,k)*conj(B(j,k));
            }
        }
}

inline double get_ek(double kx, double ky)
{
	return (kx*kx + ky*ky);
}


int main(int argc, char * argv[])
{
    //************** MPI INIT ***************************
  	int numprocs=1, myrank=0, namelen;
    
#ifndef NO_MPI
  	char processor_name[MPI_MAX_PROCESSOR_NAME];
  	MPI_Init(&argc, &argv);
  	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  	MPI_Get_processor_name(processor_name, &namelen);
    
	//cout << "Process " << myrank << " on " << processor_name << " out of " << numprocs << " says hello." << endl;
	MPI_Barrier(MPI_COMM_WORLD);
    
	MPI_Datatype ttype  = MPI_DOUBLE;
    
#endif
	if(myrank==0) cout << "\n\tProgram running on " << numprocs << " processors." << endl;
    
	//if(myrank==0)
	//	printf("\n\n");
    
#ifndef NO_MPI
	MPI_Barrier(MPI_COMM_WORLD);
#endif

    
    ofstream outfile;
    
    if(myrank==0)
    {
        outfile.open("bands.dat");
        outfile << setprecision(8);
    }
 
    
    cmat Hmat(NORB, NORB);
    vector<double> evals(NORB);
 
    int kcounter = 0;
    
    double dkx = 2.*M_PI/NUMK;
    
    for(double kx=-M_PI; kx<=M_PI+0.5*dkx; kx+=dkx)
    {
        
        Hmat.zero();

        for(int i=0; i<NORB/2; ++i)
        {
            Hmat(2*i,2*i) = -(2.*THOP*cos(kx)+MU);
            Hmat(2*i+1,2*i+1) = (2.*THOP*cos(kx)+MU);
        }
        
        // these are the diagonal blocks
        for(int i=0; i<NORB/2; ++i)
        {
            Hmat(2*i,2*i+1) = -DELTA * sin(kx) * II;
            Hmat(2*i+1,2*i) = DELTA * sin(kx) * II;
        }
        
        // these are the off-diagonal blocks
        for(int i=0; i<NORB/2-1; ++i)
        {
            Hmat(2*i,2*i+2) = -GAMMA;
            Hmat(2*i,2*i+3) = -II*GAMMA;
            Hmat(2*i+1,2*i+3) = GAMMA;
            Hmat(2*i+1,2*i+2) = -II*GAMMA;
            Hmat(2*i+2,2*i) = -GAMMA;
            Hmat(2*i+3,2*i) = II*GAMMA;
            Hmat(2*i+3,2*i+1) = GAMMA;
            Hmat(2*i+2,2*i+1) = II*GAMMA;
        }
        
        /*
        if(myrank==0 && kcounter==50)
        {
            for(int i=0; i<NORB; ++i)
            {
                for(int j=0; j<NORB; ++j)
                    cout << Hmat(i,j) << '\t';
                
                cout << endl;
            }
        }
         */
        
        diagonalize(Hmat, evals);
        
        if(myrank==0)
        {
            outfile << kcounter << '\t' << kx;
            for(int i=0; i<NORB; ++i)
            {
                outfile << '\t' << evals[i];
            }
            outfile << endl;
        }
        kcounter ++;
        
        
    }
    

    


    
    if(myrank==0)
        cout << "\n ... done! \n"<<flush;
    
    
#ifndef NO_MPI
	MPI_Finalize();
#endif
    
    return 0;

}


